use neon::prelude::*;
use neon::types::Deferred;
use once_cell::sync::{OnceCell, Lazy};
use sqlx::{postgres::{PgPoolOptions, PgPool}, Executor, Row};
use tokio::{sync::{mpsc, oneshot}, runtime::{Runtime, Handle as TokioHandle}};

static POOL: OnceCell<PgPool> = OnceCell::new();
static RUNTIME: Lazy<OnceCell<Runtime>> = Lazy::new(|| {
    OnceCell::with_value(Runtime::new().unwrap())
});

fn handle() -> &'static TokioHandle {
    RUNTIME.get().unwrap().handle()
}

enum Message {
    Query(Deferred, String, Channel),
}

fn connect(mut cx: FunctionContext) -> JsResult<JsPromise> {

    let url: Handle<JsString> = cx.argument(0)?;
    let url: String = url.value(&mut cx);

    let (deferred, promise) = cx.promise();
    let channel = cx.channel();

    let pool = POOL.get_or_init(|| {
        match PgPoolOptions::new()
        .max_connections(1)
        .min_connections(1)
        .connect_lazy(&url) {
            Ok(pool) => pool,
            Err(err) => {
                cx.throw_error(&format!("Error connecting to PostgresML: {:?}", err)).unwrap()
            }
        }
    });

    let (tx, rx) = oneshot::channel();

    let _ = tx.send(Message::Query(deferred, "SELECT version() AS version".into(), channel));

    handle().spawn(async move {
        let msg = rx.await.unwrap();

        match msg {
            Message::Query(deferred, query, channel) => {
                let result = pool.fetch_one(&*query).await.unwrap();
                let version: String = result.try_get(0).unwrap();

                deferred.settle_with(&channel, move |mut cx| {
                    Ok(cx.string(version))
                });
            }
        }
    });

    Ok(promise)
}

fn check(mut cx: FunctionContext) -> JsResult<JsNull> {
    match POOL.get() {
        Some(_) => Ok(JsNull::new(&mut cx)),
        None => {
            cx.throw_error("Not connected to PostgresML. Did you call connect(url)?")?
        }
    }
}


#[neon::main]
fn main(mut cx: ModuleContext) -> NeonResult<()> {
    cx.export_function("connect", connect)?;
    Ok(())
}
